@@ -1,13 +1,12 @@
-{-# LANGUAGE CPP                 #-}
 {-# LANGUAGE ForeignFunctionInterface #-}
 {-# LANGUAGE OverloadedStrings   #-}
 {-# LANGUAGE ScopedTypeVariables #-}
 {-# LANGUAGE TypeApplications    #-}
 {-|
 Module      : HsLua.ObjectOrientation
-Copyright   : © 2021-2022 Albert Krewinkel
+Copyright   : © 2021-2023 Albert Krewinkel
 License     : MIT
-Maintainer  : Albert Krewinkel <tarleb+hslua@zeitkraut.de>
+Maintainer  : Albert Krewinkel <tarleb@hslua.org>
 
 This module provides types and functions to use Haskell values as
 userdata objects in Lua. These objects wrap a Haskell value and provide
@@ -19,18 +18,30 @@ objects/, and to their type as /UD type/.
 module HsLua.ObjectOrientation
   ( UDType
   , UDTypeWithList (..)
+    -- * Defining types
   , deftypeGeneric
   , deftypeGeneric'
+    -- ** Methods
   , methodGeneric
+    -- ** Properties
   , property
+  , property'
   , possibleProperty
+  , possibleProperty'
   , readonly
+  , readonly'
+    -- ** Aliases
   , alias
-  , peekUD
-  , pushUD
+    -- * Marshaling
+  , peekUDGeneric
+  , pushUDGeneric
+  , initTypeGeneric
+    -- * Type docs
+  , udDocs
+  , udTypeSpec
     -- * Helper types for building
   , Member
-  , Property
+  , Property (..)
   , Operation (..)
   , ListSpec
   , Possible (..)
@@ -38,19 +49,17 @@ module HsLua.ObjectOrientation
   , AliasIndex (..)
   ) where
 
-import Control.Monad.Except
-import Foreign.Ptr (FunPtr)
+import Control.Monad ((<$!>), forM_, void, when)
 import Data.Maybe (mapMaybe)
 import Data.Map (Map)
-#if !MIN_VERSION_base(4,12,0)
-import Data.Semigroup (Semigroup ((<>)))
-#endif
 import Data.String (IsString (..))
 import Data.Text (Text)
 import Data.Void (Void)
+import Foreign.Ptr (FunPtr)
 import HsLua.Core as Lua
 import HsLua.Marshalling
 import HsLua.ObjectOrientation.Operation
+import HsLua.Typing ( TypeDocs (..), TypeSpec (..), anyType, userdataType )
 import qualified Data.Map.Strict as Map
 import qualified HsLua.Core.Unsafe as Unsafe
 import qualified HsLua.Core.Utf8 as Utf8
@@ -132,6 +141,7 @@ data Property e a = Property
   { propertyGet :: a -> LuaE e NumResults
   , propertySet :: Maybe (StackIndex -> a -> LuaE e a)
   , propertyDescription :: Text
+  , propertyType :: TypeSpec
   }
 
 -- | Alias for a different property of this or of a nested object.
@@ -162,6 +172,19 @@ data Possible a
   = Actual a
   | Absent
 
+-- | Declares a new read- and writable typed property.
+property' :: LuaError e
+          => Name                       -- ^ property name
+          -> TypeSpec                   -- ^ property type
+          -> Text                       -- ^ property description
+          -> (Pusher e b, a -> b)       -- ^ how to get the property value
+          -> (Peeker e b, a -> b -> a)  -- ^ how to set a new property value
+          -> Member e fn a
+property' name typespec desc (push, get) (peek, set) =
+  possibleProperty' name typespec desc
+    (push, Actual . get)
+    (peek, \a b -> Actual (set a b))
+
 -- | Declares a new read- and writable property.
 property :: LuaError e
          => Name                       -- ^ property name
@@ -182,7 +205,19 @@ possibleProperty :: LuaError e
   -> (Pusher e b, a -> Possible b)      -- ^ how to get the property value
   -> (Peeker e b, a -> b -> Possible a) -- ^ how to set a new property value
   -> Member e fn a
-possibleProperty name desc (push, get) (peek, set) = MemberProperty name $
+possibleProperty name = possibleProperty' name anyType
+
+-- | Declares a new read- and writable property which is not always
+-- available.
+possibleProperty' :: LuaError e
+  => Name                               -- ^ property name
+  -> TypeSpec                           -- ^ type of the property value
+  -> Text                               -- ^ property description
+  -> (Pusher e b, a -> Possible b)      -- ^ how to get the property value
+  -> (Peeker e b, a -> b -> Possible a) -- ^ how to set a new property value
+  -> Member e fn a
+possibleProperty' name typespec desc (push, get) (peek, set) =
+  MemberProperty name $
   Property
   { propertyGet = \x -> do
       case get x of
@@ -195,6 +230,24 @@ possibleProperty name desc (push, get) (peek, set) = MemberProperty name $
         Absent   -> failLua $ "Trying to set unavailable property "
                             <> Utf8.toString (fromName name)
                             <> "."
+  , propertyType = typespec
+  , propertyDescription = desc
+  }
+
+-- | Creates a read-only object property. Attempts to set the value will
+-- cause an error.
+readonly' :: Name                 -- ^ property name
+          -> TypeSpec             -- ^ property type
+          -> Text                 -- ^ property description
+          -> (Pusher e b, a -> b) -- ^ how to get the property value
+          -> Member e fn a
+readonly' name typespec desc (push, get) = MemberProperty name $
+  Property
+  { propertyGet = \x -> do
+      push $ get x
+      return (NumResults 1)
+  , propertySet = Nothing
+  , propertyType = typespec
   , propertyDescription = desc
   }
 
@@ -204,14 +257,7 @@ readonly :: Name                 -- ^ property name
          -> Text                 -- ^ property description
          -> (Pusher e b, a -> b) -- ^ how to get the property value
          -> Member e fn a
-readonly name desc (push, get) = MemberProperty name $
-  Property
-  { propertyGet = \x -> do
-      push $ get x
-      return (NumResults 1)
-  , propertySet = Nothing
-  , propertyDescription = desc
-  }
+readonly name = readonly' name anyType
 
 -- | Define an alias for another, possibly nested, property.
 alias :: AliasIndex    -- ^ property alias
@@ -220,11 +266,41 @@ alias :: AliasIndex    -- ^ property alias
       -> Member e fn a
 alias name _desc = MemberAlias name
 
+-- | Ensures that the type has been fully initialized, i.e., that all
+-- metatables have been created and stored in the registry. Returns the
+-- name of the initialized type.
+--
+-- The @hook@ can be used to perform additional setup operations. The
+-- function is called as the last step after the type metatable has been
+-- initialized: the fully initialized metatable will be at the top of
+-- the stack at that point. Note that the hook will /not/ be called if
+-- the type's metatable already existed before this function was
+-- invoked.
+initTypeGeneric :: LuaError e
+                => (UDTypeWithList e fn a itemtype -> LuaE e ())
+                -> UDTypeWithList e fn a itemtype
+                -> LuaE e Name
+initTypeGeneric hook ty = do
+  pushUDMetatable hook ty
+  pop 1
+  return (udName ty)
+
 -- | Pushes the metatable for the given type to the Lua stack. Creates
 -- the new table afresh on the first time it is needed, and retrieves it
 -- from the registry after that.
-pushUDMetatable :: LuaError e => UDTypeWithList e fn a itemtype -> LuaE e ()
-pushUDMetatable ty = do
+--
+--
+-- A @hook@ can be used to perform additional setup operations. The
+-- function is called as the last step after the type metatable has been
+-- initialized: the fully initialized metatable will be at the top of
+-- the stack at that point. Note that the hook will /not/ be called if
+-- the type's metatable already existed before this function was
+-- invoked.
+pushUDMetatable :: LuaError e
+  => (UDTypeWithList e fn a itemtype -> LuaE e ())  -- ^ @hook@
+  -> UDTypeWithList e fn a itemtype
+  -> LuaE e ()
+pushUDMetatable hook ty = do
   created <- newudmetatable (udName ty)
   when created $ do
     add (metamethodName Index)    $ pushcfunction hslua_udindex_ptr
@@ -240,6 +316,7 @@ pushUDMetatable ty = do
       Nothing -> pure ()
       Just ((pushItem, _), _) -> do
         add "lazylisteval" $ pushHaskellFunction (lazylisteval pushItem)
+    hook ty
   where
     add :: LuaError e => Name -> LuaE e () -> LuaE e ()
     add name op = do
@@ -284,7 +361,7 @@ pushGetters ty = do
   newtable
   void $ flip Map.traverseWithKey (udProperties ty) $ \name prop -> do
     pushName name
-    pushHaskellFunction $ forcePeek (peekUD ty 1) >>= propertyGet prop
+    pushHaskellFunction $ forcePeek (peekUDGeneric ty 1) >>= propertyGet prop
     rawset (nth 3)
 
 -- | Pushes the metatable's @setters@ field table.
@@ -325,7 +402,7 @@ pushAliasIndex = \case
 pairsFunction :: forall e fn a itemtype. LuaError e
               => UDTypeWithList e fn a itemtype -> LuaE e NumResults
 pairsFunction ty = do
-  obj <- forcePeek $ peekUD ty (nthBottom 1)
+  obj <- forcePeek $ peekUDGeneric ty (nthBottom 1)
   let pushMember = \case
         MemberProperty name prop -> do
           pushName name
@@ -383,10 +460,14 @@ lazyListStateName :: Name
 lazyListStateName = "HsLua unevalled lazy list"
 
 -- | Pushes a userdata value of the given type.
-pushUD :: LuaError e => UDTypeWithList e fn a itemtype -> a -> LuaE e ()
-pushUD ty x = do
-  newhsuserdata x
-  pushUDMetatable ty
+pushUDGeneric :: LuaError e
+  => (UDTypeWithList e fn a itemtype -> LuaE e ()) -- ^ push docs
+  -> UDTypeWithList e fn a itemtype                -- ^ userdata type
+  -> a                                             -- ^ value to push
+  -> LuaE e ()
+pushUDGeneric pushDocs ty x = do
+  newhsuserdatauv x 1
+  pushUDMetatable pushDocs ty
   setmetatable (nth 2)
   -- add list as value in caching table
   case udListSpec ty of
@@ -394,18 +475,18 @@ pushUD ty x = do
     Just ((_, toList), _) -> do
       newtable
       pushName "__lazylist"
-      newhsuserdata (toList x)
+      newhsuserdatauv (toList x) 1
       void (newudmetatable lazyListStateName)
       setmetatable (nth 2)
       rawset (nth 3)
-      setuservalue (nth 2)
+      void (setiuservalue (nth 2) 1)
 
 -- | Retrieves a userdata value of the given type.
-peekUD :: LuaError e => UDTypeWithList e fn a itemtype -> Peeker e a
-peekUD ty idx = do
+peekUDGeneric :: LuaError e => UDTypeWithList e fn a itemtype -> Peeker e a
+peekUDGeneric ty idx = do
   let name = udName ty
   x <- reportValueOnFailure name (`fromuserdata` name) idx
-  (`lastly` pop 1) $ liftLua (getuservalue idx) >>= \case
+  (`lastly` pop 1) $ liftLua (getiuservalue idx 1) >>= \case
     TypeTable -> do
       -- set list
       xWithList <- maybe pure setList (udListSpec ty) x
@@ -465,3 +546,21 @@ setList (_pushspec, (peekItem, updateList)) x = (x `updateList`) <$!> do
                   (y:) <$!> itemsAfter (i + 1)
               else getLazyList
       itemsAfter 1
+
+--
+-- Typing
+--
+
+-- | Returns documentation for this type.
+udDocs :: UDTypeWithList e fn a itemtype
+       -> TypeDocs
+udDocs ty = TypeDocs
+  { typeDescription = mempty
+  , typeSpec = userdataType
+  , typeRegistry = Just (udName ty)
+  }
+
+-- | Type specifier for a UDType
+udTypeSpec :: UDTypeWithList e fn a itemtype
+           -> TypeSpec
+udTypeSpec = NamedType . udName
